"""Functions to parse the formula terms and prepare arrays."""

from collections import OrderedDict

import netCDF4


def get_formula_terms(var):
    """
    Return a `formula_terms` dict mapping var_names to variables.

    The input can be a netCDF variable (`var`) holding the `formula_terms`
    attribute or the attribute itself.

    """
    formula_terms = OrderedDict()
    if isinstance(var, netCDF4.Variable):
        var = var.formula_terms
    terms = [x.strip(":") for x in var.split()]
    for k, v in zip(terms[::2], terms[1::2]):
        formula_terms.update({k: v})
    return formula_terms


def get_formula_terms_variables(nc):
    """Return a list with all variables that holds the formula_terms attribute."""

    def func(v):
        return v is not None

    var = nc.get_variables_by_attributes(formula_terms=func)
    if not var:
        msg = (
            "Could not find the attribute `formula_terms` in any of the "
            "{!r} variables."
        ).format
        raise ValueError(msg(nc))
    return var


def get_formula_terms_dims(nc, formula_terms):
    """Return an OrderedDict object `dims` holding the `formula_terms` dimensions."""
    dims = OrderedDict()
    for k, v in formula_terms.items():
        dims.update({k: nc[v].dimensions})
    return dims


def z_shape(nc, dims):
    """Return the vertical coordinate `shape` based on the formula_terms `dims`."""
    all_dims = (
        dims.get("eta"),
        dims.get("depth"),
        dims.get("sigma"),
        dims.get("s"),
    )
    all_dims = _filter_none(all_dims)
    all_dims = _flatten(all_dims)
    all_dims = _remove_duplicate(all_dims)
    if len(all_dims) > 1:
        z_dims = all_dims[:]
        z_dims.insert(1, z_dims.pop(-1))

    shape = []
    for dim in z_dims:
        try:
            shape.append(len(nc.dimensions.get(dim)))
        except TypeError:
            msg = "Could not get dimension size for dim {!r}".format
            raise ValueError(msg(dim))

    return tuple(shape)


def _reshape(arr, new_shape):
    """Reshape arrays according to the expected dimensions."""
    shape = arr.shape
    dims = [k for k in shape]
    slicer = slice(None, None, None)

    def select(dim):
        return slicer if dim in dims else None

    keys = tuple(select(k) for k in new_shape)

    return arr.__getitem__(keys)


def prepare_arrays(nc, formula_terms, new_shape):
    """Prepare arrays for formulas."""
    import dask.array as da

    arrays = {}
    for term, var in formula_terms.items():
        var = nc[var]
        if var.ndim == 0:
            arr = var[:]
        else:
            if var.ndim > 2:
                chunks = (1,) + var.shape[1:]
            else:
                chunks = var.shape
            if term == "sigma" and var.ndim == 2:
                chunks = var.shape
            if term == "eta" and var.ndim == 2:
                chunks = (1,) + var.shape[1:]
            arr = da.from_array(var, chunks=chunks)
            arr = _reshape(arr, new_shape)
        arrays.update({term: arr})
    return arrays


def _filter_none(seq):
    """Filter out None in a list."""
    return [x for x in seq if x is not None]


def _flatten(seq):
    """Flatten a list."""
    return [item for sublista in seq for item in sublista]


def _remove_duplicate(seq):
    """Drop duplicated from a list."""
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]



"""Test ocean_s_coordinate_g1."""

from pathlib import Path

import dask.array as da
import numpy as np
import pytest
from netCDF4 import Dataset

from odvc import ocean_s_coordinate_g1
from odvc.parse_formula_terms import (
    get_formula_terms,
    get_formula_terms_dims,
    get_formula_terms_variables,
    prepare_arrays,
    z_shape,
)

data_path = Path(__file__).parent.resolve().joinpath("data")


@pytest.fixture
def setup():
    """Set all the formula terms objects up for the tests."""
    roms = "ocean_s_coordinate_g1_roms.nc"
    with Dataset(data_path.joinpath(roms)) as nc:
        formula_terms_variable = get_formula_terms_variables(nc)[0]
        formula_terms = get_formula_terms(formula_terms_variable)
        dims = get_formula_terms_dims(nc, formula_terms)
        new_shape = z_shape(nc, dims)
        arrays = prepare_arrays(nc, formula_terms, new_shape)

        s = arrays["s"]
        c = arrays["C"]
        eta = arrays["eta"]
        depth = arrays["depth"]
        depth_c = arrays["depth_c"]

        z = ocean_s_coordinate_g1(s, c, eta, depth, depth_c)
        sliced = z[0, :, 30, 80]
        yield {"z": z, "sliced": sliced}


def test_shape(setup):
    """Assert that the calculated shape is correct."""
    expected = (1, 36, 82, 130)
    assert setup["z"].shape == expected


def test_slice(setup):
    """Assert that the slice is a dask array."""
    assert isinstance(setup["sliced"], da.Array)


def test_slice_ndarray(setup):
    """Assert that theslice is a numpy array."""
    assert isinstance(setup["sliced"].compute(), np.ndarray)


def test_z_values(setup):
    """Assert that the calculated z values are correct."""
    z_comp = np.array(
        [
            -2531.46656205,
            -2337.58793694,
            -2167.67949235,
            -2018.28088531,
            -1886.27977235,
            -1768.84270488,
            -1663.35273478,
            -1567.35533466,
            -1478.516278,
            -1394.59779732,
            -1313.46227081,
            -1233.11474357,
            -1151.79449631,
            -1068.11831429,
            -981.26114252,
            -891.13435155,
            -798.4970414,
            -704.92970128,
            -612.62956129,
            -524.05182095,
            -441.49011516,
            -366.71960543,
            -300.79609284,
            -244.03541034,
            -196.13254143,
            -156.34993361,
            -123.71062861,
            -97.1566744,
            -75.65879794,
            -58.28034096,
            -44.20608553,
            -32.74776123,
            -23.33605136,
            -15.50607592,
            -8.88076055,
            -3.15458049,
        ],
    )
    np.testing.assert_allclose(setup["sliced"].compute(), z_comp)



"""Test parse formula terms."""

from pathlib import Path

import dask.array as da
import numpy as np
import pytest
from netCDF4 import Dataset

from odvc.parse_formula_terms import (
    get_formula_terms,
    get_formula_terms_dims,
    get_formula_terms_variables,
    prepare_arrays,
    z_shape,
)

data_path = Path(__file__).parent.resolve().joinpath("data")


@pytest.fixture
def setup():
    """Set all the formula terms and arrays objects up for the tests."""
    roms = "ocean_s_coordinate_g1_roms.nc"
    with Dataset(data_path.joinpath(roms)) as nc:
        formula_terms_variable = get_formula_terms_variables(nc)[0]
        formula_terms = get_formula_terms(formula_terms_variable)
        dims = get_formula_terms_dims(nc, formula_terms)
        new_shape = z_shape(nc, dims)
        arrays = prepare_arrays(nc, formula_terms, new_shape)
        yield {
            "formula_terms_variable": formula_terms_variable,
            "formula_terms": formula_terms,
            "dims": dims,
            "new_shape": new_shape,
            "arrays": arrays,
        }


def test_formula_terms_variables(setup):
    """Assert that the formula_terms_variables are correctly identified."""
    assert hasattr(setup["formula_terms_variable"], "formula_terms")
    assert setup["formula_terms_variable"].standard_name == "ocean_s_coordinate_g1"
    for k, v in setup["formula_terms"].items():
        assert isinstance(v, str)


def test_get_formula_terms_and_dims(setup):
    """Assert that the formula_terms are correctly identified."""
    assert isinstance(setup["formula_terms"], dict)
    s = set(setup["formula_terms"].keys())
    assert s.issubset(["s", "C", "eta", "depth", "depth_c"])

    dims = {"s": 1, "C": 1, "eta": 3, "depth": 2, "depth_c": 0}
    for k, v in setup["dims"].items():
        assert dims[k] == len(v)


def test_arrays(setup):
    """Assert that the arrays are the right obj instance."""
    for k, v in setup["arrays"].items():
        if hasattr(v, "compute"):
            assert isinstance(v, da.Array)
            assert isinstance(v.compute(), np.ndarray)
        else:
            assert isinstance(v, np.ndarray)


def test_no_formula_terms_variables():
    """Test if it will fail as expected when formula terms are missing."""
    no_formula_term = "no_formula_terms.nc"
    with Dataset(data_path.joinpath(no_formula_term)) as nc:
        with pytest.raises(ValueError):
            get_formula_terms_variables(nc)



"""Ocean Dimensionless Vertical Coordinates."""

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"

from .formulas import (
    ocean_double_sigma_coordinate,
    ocean_s_coordinate,
    ocean_s_coordinate_g1,
    ocean_s_coordinate_g2,
    ocean_sigma_coordinate,
    ocean_sigma_z_coordinate,
)

__all__ = [
    "ocean_double_sigma_coordinate",
    "ocean_sigma_z_coordinate",
    "ocean_sigma_coordinate",
    "ocean_s_coordinate",
    "ocean_s_coordinate_g1",
    "ocean_s_coordinate_g2",
]



"""Ocean Dimensionless Vertical Coordinates formulas."""

import numpy as np


def ocean_double_sigma_coordinate(sigma, depth, z1, z2, a, href, k_c):
    """
    Create a dimensioned version of ocean double sigma.

    Definition:
      for k <= k_c
        z(k,j,i)= sigma(k)*f(j,i)
      for k > k_c
        z(k,j,i) = f(j,i) + (sigma(k)-1)*(depth(j,i)-f(j,i))
        f(j,i) = 0.5*(z1+ z2) + 0.5*(z1-z2)* tanh(2*a/(z1-z2)*(depth(j,i)-href))

    where z(k,j,i) is height, positive upwards, relative to ocean datum
    (e.g. mean sea level) at gridpoint (k,j,i) , sigma(k) is the dimensionless
    coordinate at vertical gridpoint (k) for k <= k_c , and depth(j,i) is the
    distance from ocean datum to sea floor (positive value) at horizontal
    gridpoint (j,i) . z1 , z2 , a , and href are constants.

    """
    raise NotImplementedError


def ocean_sigma_z_coordinate(sigma, eta, depth, depth_c, nsigma, zlev):
    """
    Create a dimensioned version of ocean sigma over z.

    Definition:
        for k <= nsigma
            z(n,k,j,i) = eta(n,j,i) + sigma(k)*(min(depth_c,depth(j,i))+eta(n,j,i))
        for k > nsigma
            z(n,k,j,i) = zlev(k)

        where z(n,k,j,i) is height, positive upwards, relative to ocean datum
        (e.g. mean sea level) at gridpoint (n,k,j,i) , eta(n,j,i) is the height
        of the ocean surface, positive upwards, relative to ocean datum at
        gridpoint (n,j,i) , sigma(k) is the dimensionless coordinate at
        vertical gridpoint (k) for k <= nsigma , and depth(j,i) is the distance
        from ocean datum to sea floor (positive value) at horizontal
        gridpoint (j,i) . Above depth depth_c there are nsigma layers.

    """
    raise NotImplementedError


def ocean_sigma_coordinate(sigma, eta, depth):
    """
    Create a dimensioned version of ocean sigma coordinate.

    z(n, k, j, i) = eta(n, j, i) + sigma(k) *
                    (depth(j, i) + eta(n, j, i))

    """
    return eta + sigma * (depth + eta)


def ocean_s_coordinate(s, eta, depth, a, b, depth_c):
    """
    Create a dimensioned version of ocean s-coordinate.

    z(n,k,j,i) = eta(n,j,i)*(1+s(k)) + depth_c*s(k) +
                    (depth(j,i)-depth_c)*C(k)

    where:
        C(k) = (1-b) * sinh(a*s(k)) / sinh(a) +
                b * [tanh(a * (s(k) + 0.5)) / (2 * tanh(0.5*a)) - 0.5]

    """
    c = (1 - b) * np.sinh(a * s) / np.sinh(a) + b * (
        np.tanh(a * (s + 0.5)) / (2 * np.tanh(0.5 * a)) - 0.5
    )
    return eta * (1 + s) + depth_c * s + (depth - depth_c) * c


def ocean_s_coordinate_g1(s, c, eta, depth, depth_c):
    """
    Create a dimensioned version of ocean s-coordinate generic form 1.

    z(n,k,j,i) = S(k,j,i) + eta(n,j,i) * (1 + S(k,j,i) / depth(j,i))

    where:
        S(k,j,i) = depth_c * s(k) + (depth(j,i) - depth_c) * C(k)

    """
    S = (depth_c * s) + ((depth - depth_c) * c)
    return eta * (S / depth + 1) + S


def ocean_s_coordinate_g2(s, eta, depth, depth_c, c):
    """
    Create a dimensioned version of s-coordinate generic form 2.

    z(n,k,j,i) = eta(n,j,i) + (eta(n,j,i) + depth(j,i)) * S(k,j,i)

    where:
        S(k,j,i) = (depth_c * s(k) + depth(j,i) * C(k)) /
                    (depth_c + depth(j,i))

    """
    S = (depth_c * s + depth * c) / (depth_c + depth)
    return eta + (eta + depth) * S
